import copy

from typing import Dict, Any, cast

from ..IDataTransfer import IDataTransfer
from core.constants import UnitClass, UnitProperties, UnitProperty


class Unit(IDataTransfer):

    __slots__ = [
        "__playerId",
        "__unitClass",
        "__properties"
    ]

    def __init__(self, unitClass: UnitClass = UnitClass.WORKER, playerId: int = 0):
        self.__playerId = playerId
        self.__unitClass = unitClass
        if unitClass not in UnitProperties:
            raise ValueError(f"Invalid class {unitClass}")
        self.__properties = copy.deepcopy(UnitProperties[unitClass])

    def __eq__(self, other) -> bool:
        if not isinstance(other, Unit):
            return False
        unit = cast(Unit, other)
        if self.__playerId != unit.__playerId:
            return False
        if self.__unitClass != unit.__unitClass:
            return False
        if len(self.__properties) != len(unit.__properties):
            return False
        for prop, value in self.__properties.items():
            if prop not in other.__properties:
                return False
            if value != other.__properties[prop]:
                return False
        return True

    @property
    def playerId(self) -> int:
        return self.__playerId

    @playerId.setter
    def playerId(self, playerId: int):
        self.__playerId = playerId

    @property
    def unitClass(self) -> UnitClass:
        return self.__unitClass

    def hasProperty(self, prop: UnitProperty) -> bool:
        return prop in self.__properties

    def setProperty(self, prop: UnitProperty, value: int):
        self.__properties[prop] = value

    def getProperty(self, prop: UnitProperty) -> int:
        return self.__properties[prop]

    # Data transfer

    def gatherData(self) -> Dict[str, Any]:
        properties = {}
        defaultProperties = UnitProperties[self.__unitClass]
        for prop, value in self.__properties.items():
            if value == defaultProperties[prop]:
                continue
            properties[prop.name] = value
        return {
            "playerId": self.__playerId,
            "unitClass": self.__unitClass.toName(),
            "properties": properties
        }

    def takeData(self, data: Dict[str, Any]):
        self.__playerId = int(data['playerId'])
        assert 0 <= self.__playerId < 5, f"Invalid player id {self.__playerId} in unit data"
        self.__unitClass = UnitClass.fromName(data['unitClass'])
        self.__properties = copy.deepcopy(UnitProperties[self.__unitClass])
        for name, value in data['properties'].items():
            try:
                prop = UnitProperty[name]
                self.__properties[prop] = value
            except KeyError:
                raise KeyError(f"Invalid unit property '{name}'")
