from typing import Tuple

from core.constants import UnitProperty
from core.constants.UnitClass import unitClassId2Name, UnitClass
from core.logic import Logic
from core.state import Unit
from .FrameComponent import FrameComponent
from ..control.Icon import Icon
from ..control.Label import Label
from ...theme.Theme import Theme


class UnitFrame(FrameComponent):

    def __init__(self, theme: Theme, logic: Logic, unit: Unit):
        super().__init__(theme)
        self.__logic = logic
        self.__unit = unit
        self.__unitsLayer = logic.world.units

        # Icon
        if unit is not None:
            unitClass = unit.unitClass
            playerId = unit.playerId
        else:
            unitClass = UnitClass.NONE
            playerId = 0
        unitsTileset = theme.getTileset("units")
        classTile = unitsTileset.getTile(unitClass, playerId)
        icon = Icon(theme, classTile)
        icon.moveRelativeTo("topLeft", self, "topLeft")
        self.addComponent(icon)
        
        title, actionDesc, propDesc = self.__createDescriptions()

        # Title
        titleLabel = Label(theme, title)
        titleLabel.moveRelativeTo("topLeft", icon, "topRight")
        self.addComponent(titleLabel)

        # Action points
        actionPointsLabel = Label(theme, actionDesc)
        actionPointsLabel.moveRelativeTo("topLeft", titleLabel, "bottomLeft")
        self.addComponent(actionPointsLabel)

        # Description
        descriptionLabel = Label(theme, propDesc)
        descriptionLabel.moveRelativeTo("topLeft", actionPointsLabel, "topRight")
        self.addComponent(descriptionLabel)

        self.pack()

    def __createDescriptions(self) -> Tuple[str, str, str]:
        unit = self.__unit
        if unit is None:
            return "No unit", "", ""

        title = f'{unitClassId2Name[unit.unitClass]}<flag player="{unit.playerId}">'

        actionDesc = ""
        # Life points
        if unit.hasProperty(UnitProperty.LIFE_POINTS) \
                and unit.hasProperty(UnitProperty.MAX_LIFE_POINTS):
            value = unit.getProperty(UnitProperty.LIFE_POINTS)
            maxValue = unit.getProperty(UnitProperty.MAX_LIFE_POINTS)
            actionDesc += f"{value}/{maxValue}<life><br>"
        # Move points
        if unit.hasProperty(UnitProperty.MOVE_POINTS) \
                and unit.hasProperty(UnitProperty.MAX_MOVE_POINTS):
            value = unit.getProperty(UnitProperty.MOVE_POINTS)
            maxValue = unit.getProperty(UnitProperty.MAX_MOVE_POINTS)
            actionDesc += f"{value}/{maxValue}<move><br>"
        # Attack points
        if unit.hasProperty(UnitProperty.ATTACK_POINTS) \
                and unit.hasProperty(UnitProperty.MAX_ATTACK_POINTS):
            value = unit.getProperty(UnitProperty.ATTACK_POINTS)
            maxValue = unit.getProperty(UnitProperty.MAX_ATTACK_POINTS)
            actionDesc += f"{value}/{maxValue}<attack><br>"
            
        # Attack
        propDesc = ""
        if unit.hasProperty(UnitProperty.MELEE_ATTACK):
            value = unit.getProperty(UnitProperty.MELEE_ATTACK)
            propDesc += f"{value}<melee_attack>"
        if unit.hasProperty(UnitProperty.RANGE_ATTACK):
            value = unit.getProperty(UnitProperty.RANGE_ATTACK)
            propDesc += f"{value}<range_attack>"
        if unit.hasProperty(UnitProperty.MOUNT_ATTACK):
            value = unit.getProperty(UnitProperty.MOUNT_ATTACK)
            propDesc += f"{value}<mount_attack>"
        if unit.hasProperty(UnitProperty.SIEGE_ATTACK):
            value = unit.getProperty(UnitProperty.SIEGE_ATTACK)
            propDesc += f"{value}<siege_attack>"
        propDesc += "<br>"

        # Defense
        if unit.hasProperty(UnitProperty.MELEE_DEFENSE):
            value = unit.getProperty(UnitProperty.MELEE_DEFENSE)
            propDesc += f"{value}<melee_defense>"
        if unit.hasProperty(UnitProperty.RANGE_DEFENSE):
            value = unit.getProperty(UnitProperty.RANGE_DEFENSE)
            propDesc += f"{value}<range_defense>"
        propDesc += "<br>"

        # Defense
        if unit.hasProperty(UnitProperty.MOUNT_DEFENSE):
            value = unit.getProperty(UnitProperty.MOUNT_DEFENSE)
            propDesc += f"{value}<mount_defense>"
        if unit.hasProperty(UnitProperty.SIEGE_DEFENSE):
            value = unit.getProperty(UnitProperty.SIEGE_DEFENSE)
            propDesc += f"{value}<siege_defense>"
        propDesc += "<br>"

        return title, actionDesc, propDesc

    # Component Listener

    def selectedCellChanged(self, cell: Tuple[int, int]):
        pass