from typing import Tuple, Union, Optional

import pygame

from core.constants import CellValue, UnitClass
from core.logic import Logic
from core.state import World, GameState
from tools.vector import vectorMulI, vectorSubI, vectorClampI
from .GameMode import GameMode
from ..component.IComponentListener import IComponentListener
from ..component.frame.MinimapFrame import MinimapFrame
from ..component.world.WorldComponent import WorldComponent
from ..theme.Theme import Theme


class DefaultGameMode(GameMode, IComponentListener):

    def __init__(self, theme: Theme, logic: Logic):
        super().__init__(theme)
        self._logic = logic
        self._state = state = logic.state
        self._world = world = logic.world
        self.__font = theme.getFont("default")

        # Components
        self._worldComponent = WorldComponent(theme, state)
        self._selectionLayer = self._worldComponent.selectionLayer

        self._minimapFrame = MinimapFrame(theme, world)
        self._minimapFrame.moveRelativeTo("bottomRight", self, "bottomRight")

        # Components order
        self.addComponent(self._worldComponent)
        self.addComponent(self._minimapFrame)

        # Brush
        self.__mainBrushLayer = "ground"
        self.__mainBrushValue: Union[int, str] = CellValue.GROUND_EARTH
        self.__mainBrushUnitClass: Optional[UnitClass] = None
        self.__secondaryBrushLayer = "ground"
        self.__secondaryBrushValue: Union[int, str] = CellValue.GROUND_SEA
        self.__secondaryBrushUnitClass: Optional[UnitClass] = None

        # Listener
        self._worldComponent.registerListener(self)  # worldCellClicked, ...
        self._minimapFrame.registerListener(self)  # viewChanged

    @property
    def world(self) -> World:
        return self._world

    @property
    def view(self) -> Tuple[int, int]:
        return self._worldComponent.view

    # Game loop

    def dispose(self):
        super().dispose()
        self._worldComponent.removeListener(self)
        self._minimapFrame.removeListener(self)

    def update(self):
        self._logic.executeCommands()

    # UI Event handler

    def processInput(self) -> bool:
        # Update view using keyboard state
        keys = pygame.key.get_pressed()
        newViewX, newViewY = self._worldComponent.view
        tileset = self.theme.getTileset("ground")
        tileWidth, tileHeight = tileset.tileSize
        speedX = tileWidth * 2
        speedY = tileHeight * 2
        if keys[pygame.K_RIGHT]:
            newViewX += speedX
        if keys[pygame.K_LEFT]:
            newViewX -= speedX
        if keys[pygame.K_UP]:
            newViewY -= speedY
        if keys[pygame.K_DOWN]:
            newViewY += speedY
        newView = (newViewX, newViewY)

        # Clamp new view
        worldSize = vectorMulI(self._world.size, tileset.tileSize)
        maxView = vectorSubI(worldSize, self.theme.viewSize)
        newView = vectorClampI(newView, 0, maxView)

        # Update only if changes
        if newView != self._worldComponent.view:
            self.viewChanged(newView)
            return True
        return False
