import json
import os.path
from typing import Tuple, Union, Optional

import pygame

from core.constants import CellValue, UnitClass
from core.logic import Logic
from core.state import Unit, GameState
from .DefaultGameMode import DefaultGameMode
from ..Mouse import Mouse
from ..component.frame.PaletteFrame import PaletteFrame
from ..theme.Theme import Theme


class EditGameMode(DefaultGameMode):

    def __init__(self, theme: Theme, logic: Logic):
        super().__init__(theme, logic)

        # Components
        self.__paletteFrame = PaletteFrame(theme, logic.state)
        self.__paletteFrame.moveRelativeTo("bottom", self, "bottom")

        # Components order
        self.addComponent(self.__paletteFrame)

        # Brush
        self.__mainBrushLayer = "ground"
        self.__mainBrushValue: Union[int, str] = CellValue.GROUND_EARTH
        self.__mainBrushUnitClass: Optional[UnitClass] = None
        self.__secondaryBrushLayer = "ground"
        self.__secondaryBrushValue: Union[int, str] = CellValue.GROUND_SEA
        self.__secondaryBrushUnitClass: Optional[UnitClass] = None

        # Listener
        self.__paletteFrame.registerListener(self)  # mainBrushSelected, ...

    def dispose(self):
        super().dispose()
        self.__paletteFrame.removeListener(self)

    # UI Event handler

    def keyDown(self, key: int) -> bool:
        if key in [pygame.K_e, pygame.K_F2]:
            from .PlayGameMode import PlayGameMode
            playGameMode = PlayGameMode(self.theme, self._logic)
            playGameMode.viewChanged(self.view)
            self.notifySetGameModeRequested(playGameMode)
            return True
        elif key in [pygame.K_F3]:
            try:
                self._logic.state.load("level.json")
                editGameMode = EditGameMode(self.theme, self._logic)
                editGameMode.viewChanged(self.view)
                self.notifySetGameModeRequested(editGameMode)
            except Exception as ex:
                self.notifyShowMessageRequested(str(ex), "elder")
        elif key in [pygame.K_F4]:
            try:
                self._logic.state.load("level.pickle")
                gameMode = EditGameMode(self.theme, self._logic)
                gameMode.viewChanged(self.view)
                self.notifySetGameModeRequested(gameMode)
            except Exception as ex:
                self.notifyShowMessageRequested(str(ex), "elder")
        elif key in [pygame.K_F5]:
            try:
                self._logic.state.save("level.json")
            except Exception as ex:
                self.notifyShowMessageRequested(str(ex), "elder")
        elif key in [pygame.K_F6]:
            try:
                self._logic.state.save("level.pickle")
            except Exception as ex:
                self.notifyShowMessageRequested(str(ex), "elder")
        return False

    # Component listener

    def __updateCell(self, cell: Tuple[int, int], mouse: Mouse):
        if mouse.button1 or mouse.button2:
            brushLayer = self.__mainBrushLayer
            brushValue = self.__mainBrushValue
            brushUnitClass = self.__mainBrushUnitClass
            fill = mouse.button2
        elif mouse.button3:
            brushLayer = self.__secondaryBrushLayer
            brushValue = self.__secondaryBrushValue
            brushUnitClass = self.__secondaryBrushUnitClass
            fill = False
        else:
            return
        Command = self._logic.getSetLayerCellCommand(brushLayer)
        brushUnit = None
        if brushUnitClass is not None:
            brushUnit = Unit(brushUnitClass, self._state.currentPlayerId)
        command = Command(cell, brushValue, brushUnit, fill)
        self._logic.addCommand(command)

    def worldCellClicked(self, cell: Tuple[int, int], mouse: Mouse):
        self.__updateCell(cell, mouse)

    def worldCellEntered(self, cell: Tuple[int, int], mouse: Mouse, dragging: bool):
        if dragging:
            self.__updateCell(cell, mouse)

    def mainBrushSelected(self, layerName: str, value: Union[int, str], unitClass: Optional[UnitClass] = None):
        self.__mainBrushLayer = layerName
        self.__mainBrushValue = value
        self.__mainBrushUnitClass = unitClass

    def secondaryBrushSelected(self, layerName: str, value: Union[int, str], unitClass: Optional[UnitClass] = None):
        self.__secondaryBrushLayer = layerName
        self.__secondaryBrushValue = value
        self.__secondaryBrushUnitClass = unitClass
