from typing import Tuple, Optional

import pygame

from core.logic import Logic, DistanceMap
from core.logic.command import MoveUnitCommand
from core.state import Unit
from .DefaultGameMode import DefaultGameMode
from ..Mouse import Mouse
from ui.layer.SelectionValue import SelectionValue
from ..component.frame.UnitFrame import UnitFrame
from ..theme.Theme import Theme


class PlayGameMode(DefaultGameMode):

    def __init__(self, theme: Theme, logic: Logic):
        super().__init__(theme, logic)

        # Selection
        self.__selectedCell: Optional[Tuple[int, int]] = None
        self.__selectedUnit: Optional[Unit] = None
        self.__distanceMap: Optional[DistanceMap] = None

    def dispose(self):
        super().dispose()

    # UI Event handler

    def keyDown(self, key: int) -> bool:
        if key in [pygame.K_e, pygame.K_F2]:
            from .EditGameMode import EditGameMode
            gameMode = EditGameMode(self.theme, self._logic)
            gameMode.viewChanged(self.view)
            self.notifySetGameModeRequested(gameMode)
            return True
        elif key in [pygame.K_F8]:  # Debug / Illustrate
            self._worldComponent.selectionComponent.rollShowMode()
            return True
        return False

    # Component listener

    def worldCellEntered(self, cell: Tuple[int, int], mouse: Mouse, dragging: bool):
        if self.__selectedCell is None:
            return
        if self.__distanceMap is None:
            return
        path, costs = self.__distanceMap.getPath(cell)
        self._selectionLayer.showPath(path, costs)
        self._selectionLayer.notifyContentChanged()

    def worldCellClicked(self, cell: Tuple[int, int], mouse: Mouse):
        if mouse.button1:
            if self.__selectedCell == cell:
                return
            unit = self.world.units.getUnit(cell)
            if unit is not None:
                self.__selectUnit(unit, cell)
            elif self.__selectedUnit is not None:
                self.__moveUnit(cell)
            else:
                self.__clearSelection()
        elif mouse.button3:
            self.__clearSelection()

    def __clearSelection(self):
        if self.__selectedCell is not None:
            self._selectionLayer.unselectCell(self.__selectedCell)
            self._selectionLayer.clearNumbers()
            self._selectionLayer.notifyContentChanged()
            self.__selectedCell = None

        if self.__selectedUnit is not None:
            self.__selectedUnit = None

        # Debug / Illustrate
        self._worldComponent.selectionComponent.setDistanceMap(None)

    def __selectUnit(self, unit: Unit, cell: Tuple[int, int]):
        self.__selectedCell = cell
        self.__selectedUnit = unit
        self._selectionLayer.selectCell(cell)
        self._selectionLayer.notifyCellChanged(cell)

        # Distance map
        x, y = cell
        radius = 31
        area = (
            x - radius, y - radius,
            x + radius + 1, y + radius + 1
        )
        self.__distanceMap = DistanceMap(self.world, area)
        self.__distanceMap.compute(cell)

        # Debug / Illustrate
        self._worldComponent.selectionComponent.setDistanceMap(self.__distanceMap)

    def __moveUnit(self, targetCell: Tuple[int, int]):
        if self.__selectedCell is None:
            return
        command = MoveUnitCommand(self.__selectedCell, targetCell)
        self._logic.addCommand(command)
        self.__clearSelection()
