import os
from typing import Dict, Tuple, Union, Optional, List

import numpy as np
import pygame
from pygame.color import Color
from pygame.font import Font
from pygame.surface import Surface

from .Tileset import Tileset
from .tiledef import impassableTiledef, objectsTiledef, frameTiledef, groundTiledef,\
    unitsTiledef, selectionTiledef
from .tiledef.SelectionTiledef import initSelectionTiledef


class Theme:
    def __init__(self):
        self.__viewSize = (1024, 768)

        # Tilesets
        self.__surfaces: Dict[str, Surface] = {}
        self.__tilesets: Dict[str, Tileset] = {}
        for name, tilesDef in tilesDefs.items():
            tileset = Tileset(self, tilesDef["tileSize"], tilesDef["imageFile"])
            tileset.addTiles(tilesDef["tiles"])
            self.__tilesets[name] = tileset

        self.__frameBorderSize = self.__tilesets["frame"].tileSize[0]
        self.__playerColors = [
            (128, 128, 128, 255),
            (  0,   0, 255, 255),
            (255,   0,   0, 255),
            (255, 255,   0, 255),
            (  64,255,  64, 255),
        ]

        # Text
        self.__fontsDef = {
            "default": {
                "file": "font/pixeltype/Pixeltype.ttf",
                "size": 16,
                "crop": None
            },
            "small": {
                "file": "font/lilliputsteps/lilliputsteps.ttf",
                "size": 8,
                "crop": (3, 8)
            }
        }
        self.__colors = {
            "background": Color(180, 130, 90)
        }
        self.__fontColors = {
            "default": Color(100, 75, 50),
            "warning": Color(200, 50, 35),
            "white": Color(255, 255, 255)
        }
        self.__fonts = {}

    @property
    def frameBorderSize(self) -> int:
        return self.__frameBorderSize

    @property
    def viewSize(self) -> Tuple[int, int]:
        return self.__viewSize

    @viewSize.setter
    def viewSize(self, size: Tuple[int, int]):
        self.__viewSize = size

    @property
    def playerColors(self) -> List[Tuple[int, int, int, int]]:
        return self.__playerColors

    @property
    def backgroundColor(self) -> Color:
        return self.__colors["background"]

    def init(self):
        # Add numbers to selection tileset
        initSelectionTiledef(self.__tilesets["selection"])

    def createRandomGenerator(self, seedStr: str) -> np.random.Generator:
        seed = sum(map(ord, seedStr))
        return np.random.default_rng(seed)

    def getSurface(self, imageFile: str) -> Surface:
        if imageFile not in self.__surfaces:
            fullPath = os.path.join("assets", imageFile)
            if not os.path.exists(fullPath):
                raise ValueError(f"No file '{fullPath}'")
            self.__surfaces[imageFile] = pygame.image.load(fullPath).convert()
            self.__surfaces[imageFile].set_colorkey((0, 0, 0))
        return self.__surfaces[imageFile]

    def getTileset(self, name: str) -> Tileset:
        if name not in self.__tilesets:
            raise ValueError(f"No tileset {name}")
        return self.__tilesets[name]

    def getFont(self, name: Optional[str] = None) -> Font:
        if name is None:
            name = "default"
        if name not in self.__fontsDef:
            raise ValueError("No font {}".format(name))
        fontDef = self.__fontsDef[name]
        file = os.path.join("assets", fontDef["file"])
        size = fontDef["size"]
        fontId = file, size
        if fontId not in self.__fonts:
            self.__fonts[fontId] = pygame.font.Font(file, size)
        return self.__fonts[fontId]

    def getFontCrop(self, name: Optional[str] = None) -> Union[None, Tuple[int, int]]:
        if name is None:
            name = "default"
        if name not in self.__fontsDef:
            raise ValueError("No font {}".format(name))
        fontDef = self.__fontsDef[name]
        return fontDef["crop"]

    def getFontColor(self, name: Optional[str] = None) -> Color:
        if name is None:
            name = "default"
        if name not in self.__fontColors:
            raise ValueError("No font color {}".format(name))
        return self.__fontColors[name]


tilesDefs = {
    "ground": groundTiledef,
    "impassable": impassableTiledef,
    "objects": objectsTiledef,
    "units": unitsTiledef,
    "frame": frameTiledef,
    "selection": selectionTiledef,
}

